from setuptools import setup

setup(
    name='naga',
    packages=['naga'],  # this must be the same as the name above
    version='0.2.01',
    description='A powerful Python/Lisp-hybrid-monster!',
    author='James J. Tolton',
    author_email='jjtolton@gmail.com',
    url='https://github.com/hiimjayhireme/naga',
    keywords=['testing', 'logging', 'example'],
    classifiers=[],
)
